/*
* SACD Decoder plugin
* Copyright (c) 2011-2016 Maxim V.Anisiutkin <maxim.anisiutkin@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/
 
#include "sacd_wavpack.h"

static int32_t wavpack_read_bytes(void* id, void* data, int32_t bcount) {
	return (int32_t)reinterpret_cast<sacd_media_t*>(id)->read(data, bcount);
}

static int64_t wavpack_get_pos(void* id) {
	return reinterpret_cast<sacd_media_t*>(id)->get_position();
}

static int wavpack_set_pos_abs(void*id, int64_t pos) {
	return reinterpret_cast<sacd_media_t*>(id)->seek(pos, file::seek_from_beginning) ? 0 : -1;
}

static int wavpack_set_pos_rel(void* id, int64_t delta, int mode) {
	return reinterpret_cast<sacd_media_t*>(id)->seek(delta, static_cast<file::t_seek_mode>(mode)) ? 0 : -1;
}

static int wavpack_push_back_byte(void* id, int c) {
	return reinterpret_cast<sacd_media_t*>(id)->seek(-1, file::seek_from_current) ? c : EOF;
}

static int64_t wavpack_get_length(void* id) {
	return reinterpret_cast<sacd_media_t*>(id)->get_size();
}

static int wavpack_can_seek(void* id) {
	return reinterpret_cast<sacd_media_t*>(id)->can_seek() ? 1 : 0;
}

static int32_t wavpack_write_bytes(void* id, void* data, int32_t bcount) {
	return (int32_t)reinterpret_cast<sacd_media_t*>(id)->write(data, (size_t)bcount);
}

static int wavpack_truncate_here(void* id) {
	reinterpret_cast<sacd_media_t*>(id)->truncate(reinterpret_cast<sacd_media_t*>(id)->get_position());
	return 0;
}

static int wavpack_close_stream(void* id) {
	return 0;
}

static WavpackStreamReader64 wavpack_reader = {
	wavpack_read_bytes,
	wavpack_write_bytes,
	wavpack_get_pos,
	wavpack_set_pos_abs,
	wavpack_set_pos_rel,
	wavpack_push_back_byte,
	wavpack_get_length,
	wavpack_can_seek,
	wavpack_truncate_here,
	wavpack_close_stream
};

bool g_dsd_in_wavpack(const char* file) {
	bool is_dsd = false;
	string_formatter(wavpack_file);
	wavpack_file = file;
	if (wavpack_file.find_first("file://") == 0) {
		wavpack_file.remove_chars(0, 7);
	}
	char error[80];
	int flags = OPEN_WVC | OPEN_FILE_UTF8 | OPEN_DSD_NATIVE | OPEN_ALT_TYPES;
	WavpackContext* wpc = WavpackOpenFileInput(wavpack_file, error, flags, 0);
	if (wpc) {
		is_dsd = (WavpackGetQualifyMode(wpc) & QMODE_DSD_AUDIO) ? true : false;
		WavpackCloseFile(wpc);
	}
	return is_dsd;
}

sacd_wavpack_t::sacd_wavpack_t() {
	m_file = nullptr;
	m_dB_gain_adjust = 0;
	m_wpc = nullptr;
	m_version = 0;
	m_samplerate = 0;
	m_channel_count = 0;
	m_channel_mask = 0;
	m_loudspeaker_config = 0;
	m_framerate = 0;
	m_samples_in_stream = 0;
}

sacd_wavpack_t::~sacd_wavpack_t() {
	close();
}

sacd_media_t* sacd_wavpack_t::get_media() {
	return m_file;
}

uint32_t sacd_wavpack_t::get_track_count(area_id_e area_id) {
	if (area_id == AREA_TWOCH && m_channel_count <= 2 || area_id == AREA_MULCH && m_channel_count > 2 || area_id == AREA_BOTH) {
		return 1;
	}
	return 0;
}

int sacd_wavpack_t::get_channels() {
	return m_channel_count;
}

int sacd_wavpack_t::get_loudspeaker_config() {
	return m_loudspeaker_config;
}

int sacd_wavpack_t::get_samplerate() {
	return m_samplerate;
}

int sacd_wavpack_t::get_framerate() {
	return m_framerate;
}

uint64_t sacd_wavpack_t::get_size() {
	return m_file->get_size();
}

uint64_t sacd_wavpack_t::get_offset() {
	return m_file->get_position();
}

double sacd_wavpack_t::get_duration() {
	return get_duration(0);
}

double sacd_wavpack_t::get_duration(uint32_t subsong) {
	double duration = 0.0;
	if (subsong < 1) {
		duration = (double)WavpackGetNumSamples64(m_wpc) / (double)WavpackGetSampleRate(m_wpc);
	}
	return duration;
}

bool sacd_wavpack_t::is_dst() {
	return false;
}

bool sacd_wavpack_t::open(sacd_media_t* p_file, uint32_t mode, float dB_vol_adjust) {
	m_file = p_file;
	m_dB_gain_adjust = dB_vol_adjust;
	char error[80];
	int flags = OPEN_WVC | OPEN_TAGS | OPEN_WRAPPER | OPEN_FILE_UTF8 | OPEN_DSD_NATIVE | OPEN_ALT_TYPES;
	if (m_file->get_reason() == input_open_info_write) {
		flags |= OPEN_EDIT_TAGS;
	}
	m_wpc = WavpackOpenFileInputEx64(&wavpack_reader, m_file, nullptr, error, flags, 0);
	if (!m_wpc) {
		console::printf("Error: sacd_wavpack_t::open() => %s", error);
		return false;
	}
	int qmode = WavpackGetQualifyMode(m_wpc);
	if (!(qmode & QMODE_DSD_AUDIO)) {
		throw exception_io_unsupported_format();
	}
	m_version = WavpackGetVersion(m_wpc);
	m_samplerate = WavpackGetNativeSampleRate(m_wpc);
	m_channel_count = WavpackGetNumChannels(m_wpc);
	m_channel_mask = WavpackGetChannelMask(m_wpc);
	if (m_channel_mask == 0x00 || m_channel_mask == 0xff) {
		switch (m_channel_count) {
		case 1:
			m_loudspeaker_config = 5;
			break;
		case 2:
			m_loudspeaker_config = 0;
			break;
		case 3:
			m_loudspeaker_config = 6;
			break;
		case 4:
			m_loudspeaker_config = 1;
			break;
		case 5:
			m_loudspeaker_config = 3;
			break;
		case 6:
			m_loudspeaker_config = 4;
			break;
		default:
			m_loudspeaker_config = 65535;
			break;
		}
	}
	else {
		m_loudspeaker_config = m_channel_mask | 0x10000000;
	}
	m_framerate = 75;
	m_samples_in_stream = WavpackGetNumSamples64(m_wpc);
	WavpackSeekTrailingWrapper(m_wpc);
	import_tags(WavpackGetFileFormat(m_wpc), WavpackGetWrapperData(m_wpc), WavpackGetWrapperBytes(m_wpc));
	WavpackFreeWrapper(m_wpc);
	return true;
}

bool sacd_wavpack_t::close() {
	if (m_wpc) {
		WavpackCloseFile(m_wpc);
		m_wpc = nullptr;
	}
	return true;
}

void sacd_wavpack_t::set_area(area_id_e area_id) {
}

bool sacd_wavpack_t::set_track(uint32_t track_number, area_id_e area_id, uint32_t offset) {
	return track_number < 1 ? true : false;
}

bool sacd_wavpack_t::read_frame(uint8_t* frame_data, size_t* frame_size, frame_type_e* frame_type) {
	m_data.set_size(*frame_size);
	size_t samples = *frame_size / m_channel_count;
	samples = WavpackUnpackSamples(m_wpc, m_data.get_ptr(), samples);
	for (size_t sample = 0; sample < m_data.get_size(); sample++) {
		frame_data[sample] = m_data.get_ptr()[sample];
	}
	*frame_type = FRAME_DSD;
	*frame_size = samples * m_channel_count;
	return samples > 0;
}

bool sacd_wavpack_t::seek(double seconds) {
	int samples_in_frame = m_samplerate / 8 / m_framerate;
	int64_t offset = min((int64_t)(m_samples_in_stream * (seconds / get_duration())), m_samples_in_stream);
	return WavpackSeekSample64(m_wpc, (offset / samples_in_frame) * samples_in_frame) == TRUE;
}

void sacd_wavpack_t::get_info(uint32_t subsong, file_info& info) {
	array_t<char> tag_name;
	array_t<char> tag_value;
	replaygain_info rg_info = info.get_replaygain();
	int tag_items = WavpackGetNumTagItems(m_wpc);
	if (tag_items == 0) {
		if (m_tags.get_size() > 0) {
			try {
				service_ptr_t<file> f;
				abort_callback_impl abort;
				filesystem::g_open_tempmem(f, abort);
				f->write(m_tags.get_ptr(), m_tags.get_size(), abort);
				tag_processor::read_id3v2(f, info, abort);
			}
			catch (exception_io_data) {
			}
		}
	}
	for (int tag_index = 0; tag_index < tag_items; tag_index++) {
		tag_name.set_size(WavpackGetTagItemIndexed(m_wpc, tag_index, nullptr, 0) + 1);
		WavpackGetTagItemIndexed(m_wpc, tag_index, tag_name.get_ptr(), tag_name.get_size());
		tag_value.set_size(WavpackGetTagItem(m_wpc, tag_name.get_ptr(), nullptr, 0) + 1);
		WavpackGetTagItem(m_wpc, tag_name.get_ptr(), tag_value.get_ptr(), tag_value.get_size());
		bool rg_info_set = false;
		if (strcmp(tag_name.get_ptr(), "replaygain_track_gain") == 0) {
			rg_info.set_track_gain_text(tag_value.get_ptr());
			rg_info_set = true;
		}
		if (strcmp(tag_name.get_ptr(), "replaygain_track_peak") == 0) {
			rg_info.set_track_peak_text(tag_value.get_ptr());
			rg_info_set = true;
		}
		if (strcmp(tag_name.get_ptr(), "replaygain_album_gain") == 0) {
			rg_info.set_album_gain_text(tag_value.get_ptr());
			rg_info_set = true;
		}
		if (strcmp(tag_name.get_ptr(), "replaygain_album_peak") == 0) {
			rg_info.set_album_peak_text(tag_value.get_ptr());
			rg_info_set = true;
		}
		if (!rg_info_set) {
			info.meta_set(tag_name.get_ptr(), tag_value.get_ptr());
		}
	}
	adjust_replaygain(rg_info, m_dB_gain_adjust);
	info.set_replaygain(rg_info);
}

void sacd_wavpack_t::set_info(uint32_t subsong, const file_info& info) {
	t_size meta_tags = info.meta_get_count();
	for (t_size tag_index = 0; tag_index < info.meta_get_count(); tag_index++) {
		const char* tag_name = info.meta_enum_name(tag_index);
		for (t_size value_index = 0; value_index < info.meta_enum_value_count(tag_index); value_index++) {
			const char* tag_value = info.meta_enum_value(tag_index, value_index);
			WavpackAppendTagItem(m_wpc, tag_name, tag_value, strlen(tag_value));
		}
	}
	replaygain_info rg_info = info.get_replaygain();
	adjust_replaygain(rg_info, -m_dB_gain_adjust);
	replaygain_info::t_text_buffer rg_value;
	if (rg_info.is_track_gain_present()) {
		if (rg_info.format_track_gain(rg_value)) {
			WavpackAppendTagItem(m_wpc, "replaygain_track_gain", rg_value, strlen(rg_value));
		}
	}
	if (rg_info.is_track_peak_present()) {
		if (rg_info.format_track_peak(rg_value)) {
			WavpackAppendTagItem(m_wpc, "replaygain_track_peak", rg_value, strlen(rg_value));
		}
	}
	if (rg_info.is_album_gain_present()) {
		if (rg_info.format_album_gain(rg_value)) {
			WavpackAppendTagItem(m_wpc, "replaygain_album_gain", rg_value, strlen(rg_value));
		}
	}
	if (rg_info.is_album_peak_present()) {
		if (rg_info.format_album_peak(rg_value)) {
			WavpackAppendTagItem(m_wpc, "replaygain_album_peak", rg_value, strlen(rg_value));
		}
	}
}

bool sacd_wavpack_t::commit() {
	return WavpackWriteTag(m_wpc) == TRUE;
}

bool sacd_wavpack_t::set_samplerate(sacd_media_t* p_file, uint32_t p_samplerate) {
	return false;
}

void sacd_wavpack_t::adjust_replaygain(replaygain_info& rg_info, float dB_vol_adjust) {
	float adj_coef = pow(10.0f, dB_vol_adjust / 20.0f);
	if (rg_info.is_album_gain_present()) {
		rg_info.m_album_gain -= dB_vol_adjust;
	}
	if (rg_info.is_album_peak_present()) {
		rg_info.m_album_peak *= adj_coef;
	}
	if (rg_info.is_track_gain_present()) {
		rg_info.m_track_gain -= dB_vol_adjust;
	}
	if (rg_info.is_track_peak_present()) {
		rg_info.m_track_peak *= adj_coef;
	}
}

void sacd_wavpack_t::import_tags(uint8_t wrapper_type, uint8_t* wrapper_data, uint32_t wrapper_size) {
	Chunk ck;
	ID id;
	m_tags.set_size(0);
	int offset = 0;
	switch (wrapper_type) {
	case WP_FORMAT_DFF:
		memcpy(&ck, wrapper_data + offset, sizeof(ck));
		if (!ck.has_id("FRM8")) {
			return;
		}
		offset += sizeof(ck);
		memcpy(&id, wrapper_data + offset, sizeof(id));
		if (!id.has_id("DSD ")) {
			return;
		}
		offset += sizeof(id);
		break;
	case WP_FORMAT_DSF:
		memcpy(&ck, wrapper_data + offset, sizeof(ck));
		if (!ck.has_id("DSD ")) {
			return;
		}
		offset += sizeof(ck);
		offset += 8;
		offset += 8;
		break;
	}
	while (offset + sizeof(ck) < wrapper_size) {
		switch (wrapper_type) {
		case WP_FORMAT_DFF:
			memcpy(&ck, wrapper_data + offset, sizeof(ck));
			if (ck.has_id("ID3 ")) {
				m_tags.append_fromptr(wrapper_data + offset + sizeof(ck), (t_size)ck.get_size());
			}
			offset += sizeof(ck);
			if (!ck.has_id("DSD ")) {
				offset += (int)ck.get_size();
			}
			offset += offset & 1;
			break;
		case WP_FORMAT_DSF:
			memcpy(&ck, wrapper_data + offset, sizeof(ck));
			if (ck.has_id("data")) {
				offset += sizeof(ck);
				m_tags.append_fromptr(wrapper_data + offset, wrapper_size - offset);
				offset = wrapper_size;
			}
			else {
				offset += (int)ck.ckDataSize;
			}
			break;
		}
	}
}
